/*
 * mktape
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <syslog.h>
#include <string.h>
#include <time.h>
#include <inttypes.h>
#include <pwd.h>
#include "be_byteshift.h"
#include "mhvtl_list.h"
#include "vtl_common.h"
#include "vtltape.h"
#include "vtllib.h"

#if defined _LARGEFILE64_SOURCE
static void *largefile_support = "large file support";
#else
static void *largefile_support = "No largefile support";
#endif

/* The following variables are needed for the MHVTL_DBG() macro to work. */

char mhvtl_driver_name[] = "mktape";
int verbose = 0;
int debug = 0;
long my_id = 0;
int cdbonly = 0;

extern char home_directory[HOME_DIR_PATH_SZ + 1];

static void usage(char *progname)
{
  printf("Usage: %s -l lib -c cont -m PCL -n no_of_tapes -s size -t type -d density\n",
	 progname);
	printf("Where OPTIONS are from:\n");
	printf("      -h    -- print this message and exit\n");
	printf("      -v    -- verbose\n");
	printf("      -D[N] -- set debug level to N [1]\n");
	printf("      -V    -- print Version and exit\n");
	printf("And REQUIRED-PARAMS are:\n");
	printf("      -l lib          -- set Library number\n");
	printf("      -m PCL          -- set Physical Cartrige Label (barcode)\n");
	printf("      -s size         -- set size in Megabytes\n");
	printf("      -t type         -- set to data, clean, WORM, or NULL\n");
	printf("      -b tape-base    -- set tape base\n");
	printf("      -c name         -- set container name\n");
	printf("      -n no_of_tapes  -- set number of tapes to be created");
	printf("      -d density      -- set density to one of the following:\n");
	printf("           AIT1     AIT2     AIT3     AIT4\n");
	printf("           DDS1     DDS2     DDS3     DDS4\n");
	printf("           DLT3     DLT4\n");
	printf("           SDLT1    SDLT220  SDLT320  SDLT600\n");
	printf("           LTO1     LTO2     LTO3     LTO4\n");
	printf("           LTO5     LTO6     LTO7     LTO8\n");
	printf("           T10KA    T10KB    T10KC\n");
	printf("           9840A    9840B    9840C    9840D\n");
	printf("           9940A    9940B\n");
	printf("           J1A      E05      E06      E07\n");
	printf("\n");
}

static char *__get_base36(int base10, char *alphanum)
{
	char *base36char = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	int i = 0;
	for(i=3-1; i>=0; --i) {
		alphanum[i] = base36char[base10 % 36];
		base10 /= 36;
	}
	return alphanum;
}

int gettime()
{
	const char *source_date_epoch = getenv("SOURCE_DATE_EPOCH");
	time_t t;
	if ( source_date_epoch == NULL || (t = (time_t)strtoll(source_date_epoch, NULL, 10)) <= 0)
		t = time(NULL);
	return t;
}

int main(int argc, char *argv[])
{
	unsigned char sam_stat;
	char *progname = argv[0];
	int pcl = 0, i,rc, err = 0;
	char *mediaType = NULL;
	char *mediaCapacity = NULL;
	char *density = NULL;
	char *lib = NULL;
	char *tape_base = NULL;
	char *cont_name = NULL;
	uint64_t size;
	int libno;
	int no_of_tapes;
	struct passwd *pw;
	struct stat statb;
	char spcl[10];
	char home_dir[64];
	char alphanum[3];
	int opt;
	int res;
	
	if (sizeof(struct MAM) != 1024) {
		fprintf(stderr, "error: Structure of MAM incorrect size: %d\n",
						(int)sizeof(struct MAM));
		exit(2);
	}

	if (argc < 2) {
		fprintf(stderr, "error: not enough arguments\n");
		usage(progname);
		exit(1);
	}

	while ((opt = getopt(argc, argv, "d:l:m:s:t:VvD::hC:H:")) != -1) {
		switch (opt) {
		case 'h':
			usage(progname);
			exit(0);
		case 'd':
			density = strdup(optarg);
			break;
		case 'l':
			lib = strdup(optarg);
			break;
		case 'm':
			pcl = atoi(optarg);
			break;
		case 's':
			mediaCapacity = strdup(optarg);
			break;
		case 't':
			mediaType = strdup(optarg);
			break;
		case 'b':
			tape_base = strdup(optarg);
			break;
		case 'c':
			cont_name = strdup(optarg);
			break;
		case 'n':
			no_of_tapes = atoi(optarg);
			break;
		case 'V':
			printf("%s: version %s\n", progname, MHVTL_VERSION);
			printf("%s\n", (char *)largefile_support);
			exit(0);
		case 'D':
			if (optarg)
				debug = strtol(optarg, NULL, 0);
			else
				debug++;
			break;
		case 'v':
			verbose++;
			break;
		default:
			fprintf(stderr, "error: unknown option: %c\n", opt);
			usage(progname);
			exit(1);
		}
	}

	if ((pcl <= 0 ) || (pcl >= 1295)){
		printf("Please supply a barcode (-m barcode) between 1 and 1295\n\n");
		usage(progname);
		exit(1);
	}
	
	if (mediaCapacity == NULL) {
		fprintf(stderr, "error: Please supply media capacity (-s xx)\n\n");
		usage(progname);
		exit(1);
	}
	if (mediaType == NULL) {
		fprintf(stderr, "error: Please supply cart type (-t data|clean|WORM|NULL)\n\n");
		usage(progname);
		exit(1);
	}
	if (lib == NULL) {
		fprintf(stderr, "error: Please supply Library number (-l xx)\n\n");
		usage(progname);
		exit(1);
	}

	if (tape_base == NULL) {
		printf("Please supply Tape Base (-b xxxxxxx)\n\n");
		usage(progname);
		exit(1);
	}
	
	if (density == NULL) {
		fprintf(stderr, "error: Please supply media density (-d xx)\n\n");
		usage(progname);
		exit(1);
	}

	GET_CONTAINER_ROOT(home_dir, cont_name);
	if (stat(home_dir, &statb) < 0 && errno == ENOENT) {
		printf("Please supply a valid container name\n\n");
		usage(progname);
		exit(1);
	} 

	sscanf(mediaCapacity, "%" PRId64, &size);
	if (size == 0)
		size = 8000;

	sscanf(lib, "%d", &libno);
	if (!libno) {
		fprintf(stderr, "error: Invalid library number\n");
		exit(1);
	}

	find_media_home_directory(NULL, home_directory, libno, cont_name);

	pw = getpwnam(USR);     /* Find UID for user 'vtl' */
	
	/* Verify that the MHVTL home directory exists. */
	if (stat(home_directory, &statb) < 0 && errno == ENOENT) {
	  umask(0007);
	  if (mkdir(home_directory, 02770) < 0) {
	    printf("Cannot create PCL %d, directory %s:"
		   "Doesn't exist and cannot be created\n",
		   pcl, home_directory);
	    exit(1);
	  }
	}

	/* Don't really care if this fails or not..
	 * But lets try anyway
	 */
	if (chown(home_directory, pw->pw_uid, pw->pw_gid)) {
	}

	vtl_log_init();
	
	/* Initialize the contents of the MAM to be used for the new PCL. */
	for (i = pcl; i < (pcl+no_of_tapes); i++) {
	  __get_base36(i, alphanum);
	  snprintf(spcl, sizeof(spcl), "%s%s", tape_base, alphanum);
	  memset((uint8_t *)&mam, 0, sizeof(mam));
	  mam.tape_fmt_version = TAPE_FMT_VERSION;
	  mam.mam_fmt_version = MAM_VERSION;
	  put_unaligned_be64(size * 1048576, &mam.max_capacity);
	  put_unaligned_be64(size * 1048576, &mam.remaining_capacity);
	  put_unaligned_be64(sizeof(mam.pad), &mam.MAMSpaceRemaining);

	  memcpy(&mam.MediumManufacturer, "linuxVTL", 8);
	  memcpy(&mam.ApplicationVendor, "vtl-1.7 ", 8);
	  sprintf((char *)mam.ApplicationVersion, "%d", TAPE_FMT_VERSION);

	  if (!strncmp("clean", mediaType, 5)) {
	    mam.MediumType = MEDIA_TYPE_CLEAN;	/* Cleaning cart */
	    mam.MediumTypeInformation = 20;		/* Max cleaning loads */
	  } else if (!strncmp("NULL", mediaType, 4)) {
	    mam.MediumType = MEDIA_TYPE_NULL;	/* save metadata only */
	  } else if (!strncmp("WORM", mediaType, 4)) {
	    mam.MediumType = MEDIA_TYPE_WORM;	/* WORM cart */
	  } else {
	    mam.MediumType = MEDIA_TYPE_DATA;	/* Normal data cart */
	  }
	  set_media_params(&mam, density);
	  
	  sprintf((char *)mam.MediumSerialNumber, "%s_%d", spcl, (int)gettime());
	  sprintf((char *)mam.MediumManufactureDate, "%d", (int)gettime());
	  sprintf((char *)mam.Barcode, "%-31s", spcl);

	  if (verbose)
	    printf("Creating tape data ...\n");

	  /* Create the PCL using the initialized MAM. */
	  res = create_tape(spcl, &mam, &sam_stat, home_directory, &err);
	  if (res) {
	    printf("Cartridge creation failed with error, %s\n", strerror(err));
	    vtl_log_close();
	    exit(res);
	  }
	}

	vtl_log_close();
	
	exit(0);
}
