/*
 * Version 2 of tape format.
 *
 * Each media contains 3 files.
 *  - The .data file contains each block of data written to the media
 *  - The .indx file consists of an array of one raw_header structure per
 *    written tape block or filemark.
 *  - The .meta file consists of a MAM structure followed by a meta_header
 *    structure, followed by a variable-length array of filemark block numbers.
 *
 * Copyright (C) 2009 - 2010 Kevan Rehm

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _FILE_OFFSET_BITS 64

#define __STDC_FORMAT_MACROS	/* for PRId64 */

/* Limitation of underlying FS 256K */
#define MAX_IOSZ (256 * 1024)

/* for unistd.h pread/pwrite and fcntl.h posix_fadvise */
/* #define _XOPEN_SOURCE 600 */

#include <sys/syslog.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <pwd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include "logging.h"
#include "mhvtl_scsi.h"
#include "mhvtl_list.h"
#include "vtltape.h"
#include "be_byteshift.h"

#ifndef USR
#define USR "vtl"
#endif
/* 
 * Every cartridge is represented by three files:
 *
 * data: This is the real data on tape. It does not contain filemark
 *       information.
 *
 * indx: For every write (variable or FIXED) and for every FILEMARK, 
 *       we write 512 bytes in this file. It contains the data offset 
 *       in the data file where the data is available and a block header 
 *       which contains type, flags, block number and block size. For 
 *       example if we are writing a FM at a block, we record the block 
 *       type as B_FILEMARK and if writing data, B_DATA.
 *
 * meta: This contains the MAM structure followed by a meta_header followed 
 *       by an array of filemark block numbers. The MAM (Medium Auxillary Memory) 
 *       structure is 1024 bytes and has many things about the cartridge 
 *       including space remaining, load count, etc. See the Ultrium tape drive
 *       reference manual for more on MAM. Following the MAM is a 512 byte 
 *       block that contains the file mark count. Following this is an array 
 *       of 32 bit values each indicating block number where a FM is. 
 */

/* 
 * The .indx file consists of an array of one raw_header structure per
 * written tape block or filemark.  There is no separate raw_header
 * structure required for BOT or EOM.  The raw_header structure is padded
 * out to 512 bytes to allow for the addition of fields in the future without
 * breaking backwards compatibility with existing PCL indx files.
 */

struct raw_header {
	loff_t	data_offset;
	struct blk_header hdr;
	char pad[512 - sizeof(loff_t) - sizeof(struct blk_header)];
};

/* 
 * The .meta file consists of a MAM structure followed by a 
 * meta_header structure, followed by a variable-length array 
 * of filemark block numbers. Both the MAM and meta_header 
 * structures also contain padding to allow for future expansion 
 * with backwards compatibility.
 *
 * This is sector aligned for atomicity purposes.
 */
struct	meta_header {
	uint32_t filemark_count;
	uint32_t filemark_count_save;
	char pad[512 - (2 * sizeof(uint32_t))];
};

char currentbar[16];
static char currentPCL[1024];
static int datafile = -1;
static int indxfile = -1;
static int metafile = -1;
static int exportfile = -1;

static struct raw_header raw_pos;
static struct meta_header meta;
static uint64_t eod_data_offset;
static uint32_t eod_blk_number;

#define FM_DELTA 500
static int filemark_alloc = 0;
static uint32_t *filemarks = NULL;

/* Globally visible variables. */
struct indx_buffer ibuf;
struct MAM mam;
struct blk_header *c_pos = &raw_pos.hdr;
int OK_to_write = 0;
char home_directory[HOME_DIR_PATH_SZ + 1 + MAX_BARCODE_LEN];
static int vtl_readonly = 0;

static void cartsync(void)
{
	if (datafile >= 0) {
		fsync(datafile);
	}
	if (indxfile >= 0) {
       		fsync(indxfile);
	}
	if (metafile >= 0) {
       		fsync(metafile);
	}
	return;
}

static void cartclean(void)
{
	if (datafile >= 0) {
		fsync(datafile);
		close(datafile);
		datafile = -1;
	}
	if (indxfile >= 0) {
       		fsync(indxfile);
		close(indxfile);
		indxfile = -1;
	}
	if (metafile >= 0) {
       		fsync(metafile);
		close(metafile);
		metafile = -1;
	}
	if (exportfile >= 0) {
		close(exportfile);
		exportfile = -1;
	}
	return;
}

#ifdef MHVTL_DEBUG
static char *mhvtl_block_type_desc(int blk_type)
{
	unsigned int i;

	static const struct {
		int blk_type;
		char *desc;
		} block_type_desc[] = {
			{ B_FILEMARK, "FILEMARK" },
			{ B_EOD, "END OF DATA" },
			{ B_NOOP, "NO OP"},
			{ B_DATA, "DATA" },
		};
	for (i = 0; i < ARRAY_SIZE(block_type_desc); i++)
		if (block_type_desc[i].blk_type == blk_type)
			return block_type_desc[i].desc;
	return NULL;
}
#endif

/*
 * Returns:
 * == 0, success
 * != 0, failure
*/

static int mkEODHeader(uint32_t blk_number, uint64_t data_offset)
{
	memset(&raw_pos, 0, sizeof(raw_pos));

	raw_pos.data_offset = data_offset;

	raw_pos.hdr.blk_type = B_EOD;
	raw_pos.hdr.blk_number = blk_number;

	eod_blk_number = blk_number;
	eod_data_offset = data_offset;

	OK_to_write = 1;

	return 0;
}

/*
 * Returns:
 * == 0, success
 * != 0, failure
*/
int flush_index(uint8_t *sam_stat)
{
  ssize_t nwrite = 0;
  uint32_t blk_number;
  uint64_t data_offset;
  
  blk_number = raw_pos.hdr.blk_number;
  data_offset = raw_pos.data_offset;
  
  if (vtl_readonly) {
    if (sam_stat) {
      sam_data_protect(E_WRITE_PROTECT, sam_stat);
    }
    
    MHVTL_ERR("VTL in read-only mode. Index file write failure, start %" PRId64
	      ": end % " PRId64, (uint64_t)ibuf.offset,
	      (uint64_t)blk_number * sizeof(raw_pos));
    
    return -1;
  }

  /* Flush out the index buffer.*/
  if (ibuf.idx) {
    MHVTL_DBG(3, "FI: flush: ibuf.idx %d, ibuf.data_written %ld, blk_number %d",
	      ibuf.idx, ibuf.data_written, blk_number);
    
    MHVTL_DBG(3, "FI: offset %ld, size %ld", ibuf.offset, (ibuf.idx * sizeof(raw_pos)));
    
    nwrite = pwrite(indxfile, ibuf.buf, (ibuf.idx * sizeof(raw_pos)), ibuf.offset); 
    
    /*(blk_number - ibuf.idx) * sizeof(raw_pos)); */
    if (nwrite != (ssize_t)(ibuf.idx * sizeof(raw_pos))) {
      long indxsz = (((blk_number - 1) - ibuf.idx) > 0) ?
	(((blk_number - 1) - ibuf.idx) * sizeof(raw_pos)) : 0;
      
      long datasz = ((data_offset - ibuf.data_written) > 0) ?
	(data_offset - ibuf.data_written) : 0;
      
      if (sam_stat) {
	sam_medium_error(E_WRITE_ERROR, sam_stat);
      }
      
      MHVTL_ERR("Index file write failure, start pos: %" PRId64 ": end pos %" PRId64 "%s",
		(uint64_t)ibuf.offset, (uint64_t)blk_number * sizeof(raw_pos),
		strerror(errno));
      
      vtl_readonly = (errno == ENOSPC);
      
      MHVTL_DBG(3, "Truncating index file size to: %ld", indxsz);
      if (ftruncate(indxfile, indxsz) < 0) {
	MHVTL_ERR("Error truncating indx: %s", strerror(errno));
      }
      
      MHVTL_DBG(3, "Truncating data file size: %"PRId64, datasz);
      if (ftruncate(datafile, datasz) < 0) {
	MHVTL_ERR("Error truncating data: %s", strerror(errno));
      }
      
      mkEODHeader((indxsz/sizeof(raw_pos)), datasz);

      ibuf.offset = indxsz;
      ibuf.idx = ibuf.data_written = 0;

      memset(ibuf.buf, 0, INDEX_BUF_SIZE);

      return -1;
    }
    
    ibuf.offset += (ibuf.idx * sizeof(raw_pos));
    ibuf.idx = ibuf.data_written = 0;
    memset(ibuf.buf, 0, INDEX_BUF_SIZE);
    
    fsync(indxfile);
  }
  
  return 0;
}

/*
 * Returns:
 * == 0, success
 * != 0, failure
*/

static int read_header(uint32_t blk_number, uint8_t *sam_stat)
{
	loff_t nread;

	if (flush_index(sam_stat)) {
	  MHVTL_DBG(3, "read_header: flush_index returned error, rofs.");
	  /* do not return error, since it is just a read operation */
	}
	
	MHVTL_DBG(3, "Reading header for block %d", blk_number);

	if (blk_number > eod_blk_number) {
		MHVTL_ERR("Attempt to seek [%d] beyond EOD [%d]",
				blk_number, eod_blk_number);
	} else if (blk_number == eod_blk_number) {
		mkEODHeader(eod_blk_number, eod_data_offset);

		MHVTL_DBG(3, "read_header: EOD: ibuf.idx %d, ibuf.offset %ld, "
			  "blk_number %d, eod_blk_number %d", ibuf.idx, ibuf.offset,
			  blk_number, eod_blk_number);
		
		ibuf.offset = blk_number * sizeof(raw_pos);
	} else {
		nread = pread(indxfile, &raw_pos, sizeof(raw_pos),
			blk_number * sizeof(raw_pos));
		if (nread < 0) {
			MHVTL_ERR("Medium format corrupt");
			sam_medium_error(E_MEDIUM_FMT_CORRUPT, sam_stat);
			return -1;
		} else if (nread != sizeof(raw_pos)) {
			MHVTL_ERR("Failed to read next header");
			sam_medium_error(E_END_OF_DATA, sam_stat);
			return -1;
		}

		MHVTL_DBG(3, "read_header: offset: ibuf.idx %d, ibuf.offset %ld, "
			  "blk_number %d, eod_blk_number %d", ibuf.idx, ibuf.offset,
			  blk_number, eod_blk_number);
		
		ibuf.offset = blk_number * sizeof(raw_pos);
	}

	MHVTL_DBG(3, "Reading header %d at offset %ld, type: %s, size: %d",
			raw_pos.hdr.blk_number,
			(unsigned long)raw_pos.data_offset,
			mhvtl_block_type_desc(raw_pos.hdr.blk_type),
			raw_pos.hdr.blk_size);
	return 0;
}

static int tape_loaded(uint8_t *sam_stat)
{
	if (datafile != -1)
		return 1;

	sam_not_ready(E_MEDIUM_NOT_PRESENT, sam_stat);
	return 0;
}

static int rewrite_meta_file(int log_error)
{
	ssize_t io_size, nwrite, failed = 0;
	size_t io_offset;

	if (vtl_readonly) {
		MHVTL_ERR("VTL is in read-only mode ... Skipping write.");
		return -1;
	}

	if (exportfile >= 0) {
		// Cart exported to cloud, avoid meta writes
		MHVTL_LOG("Avoiding MAM rewrite as cart is exported to cloud");
		return -1;
	}

retry:
	io_size = sizeof(meta);
	io_offset = sizeof(struct MAM);
	nwrite = pwrite(metafile, &meta, io_size, io_offset);
	if (nwrite < 0) {
	  if (log_error)
	    MHVTL_ERR("Error writing meta_header to metafile: %s",
		      strerror(errno));
	  
	  vtl_readonly = (errno == ENOSPC || errno == EROFS);
	  
	  return -1;
	}
	if (nwrite != io_size) {
	  if (log_error)
		MHVTL_ERR("Error writing meta_header map to metafile."
			  " Expected to write %d bytes", (int)io_size);
	  return -1;
	}

	if (failed)
	  return -1; 

	io_size = meta.filemark_count * sizeof(*filemarks);
	io_offset = sizeof(struct MAM) + sizeof(meta);

	if (io_size) {
		nwrite = pwrite(metafile, filemarks, io_size, io_offset);
		if (nwrite < 0) {
		  if (log_error){
		    MHVTL_ERR("Error writing filemark map to metafile: %s",
			      strerror(errno));
		    MHVTL_ERR("Trying to revert meta file ...");
		  }
		  meta.filemark_count = meta.filemark_count_save;
		  vtl_readonly = (errno == ENOSPC || errno == EROFS);
		  failed = 1;
		  goto retry;
		  return -1;
		}
		if (nwrite != io_size) {
		  if (log_error)
		    MHVTL_ERR("Error writing filemark map to metafile."
			      " Expected to write %d bytes", (int)io_size);
		  return -1;
		}
	}

	/* 
	 * If filemarks were overwritten, the meta file may need to be shorter
	 * than before.
	 */
	if (ftruncate(metafile, io_offset + io_size) < 0) {
	  if (log_error)
	    MHVTL_ERR("Error truncating metafile: %s", strerror(errno));
	  vtl_readonly = (errno == ENOSPC || errno == EROFS);
	  return -1;
	}

	return 0;
}

static int check_for_overwrite(uint8_t *sam_stat)
{
	uint32_t blk_number;
	uint64_t data_offset;
	unsigned int i;
	int fd = -1, err = 0;
	char data_file_name[HOME_DIR_PATH_SZ+1] = "";
	struct passwd *pw;
	
	if (raw_pos.hdr.blk_type == B_EOD)
		return 0;

	MHVTL_DBG(2, "At block %ld", (unsigned long)raw_pos.hdr.blk_number);

	/* We aren't at EOD so we are performing a rewrite.  Truncate
	   the data and index files back to the current length.
	*/

	blk_number = raw_pos.hdr.blk_number;
	data_offset = raw_pos.data_offset;

	if (ftruncate(indxfile, blk_number * sizeof(raw_pos))) {
		sam_medium_error(E_WRITE_ERROR, sam_stat);
		MHVTL_ERR("Index file ftruncate failure, pos: "
			"%" PRId64 ": %s",
			(uint64_t)blk_number * sizeof(raw_pos),
			strerror(errno));
		vtl_readonly = (errno == ENOSPC || errno == EROFS);
		return -1;
	}
	
	MHVTL_DBG(3, "check_for_overwrite: ibuf.idx %d, ibuf.data_written %ld, "
		  "blk_number %d", ibuf.idx, ibuf.data_written, blk_number);
	
	ibuf.offset = blk_number * sizeof(raw_pos);
	
        if (data_offset == 0){
	  MHVTL_DBG(3, "check_for_overwrite: overwrite at 0, unlink and recreate "
		    "the data file %s/data instead of ftruncate", currentPCL);
	  close(datafile);
	  datafile = -1;
	  snprintf(data_file_name, ARRAY_SIZE(data_file_name), "%s/data", currentPCL);
	  err = unlink(data_file_name);
	  if (err) {
	    sam_medium_error(E_WRITE_ERROR, sam_stat);
	    MHVTL_ERR("unable to unlink %s: %s", data_file_name, strerror(errno));
	    vtl_readonly = (errno == ENOSPC || errno == EROFS);
	    return -1;
	  }
	  
	  fd = creat(data_file_name, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
	  if (fd < 0) {
	    sam_medium_error(E_WRITE_ERROR, sam_stat);
	    MHVTL_ERR("unable to create %si: %s", data_file_name, strerror(errno));
	    vtl_readonly = (errno == ENOSPC || errno == EROFS);
	    return -1;
	  }
	  close(fd);
	  pw = getpwnam(USR);     /* Find UID for user 'vtl' */
	  if (!pw) {
	    MHVTL_ERR("Failed to get UID for user '%s': %s", USR,
		      strerror(errno));
	    return 1;
	  }
	  chown(data_file_name, pw->pw_uid, pw->pw_gid);
	  datafile = open(data_file_name, O_RDWR|O_LARGEFILE);
	  if (datafile < 0){
	    MHVTL_ERR("unable to open %s: %s", data_file_name, strerror(errno));
	    sam_medium_error(E_WRITE_ERROR, sam_stat);
	    return -1;
	  }
        } else {
	  if (ftruncate(datafile, data_offset)) {
	    sam_medium_error(E_WRITE_ERROR, sam_stat);
	    MHVTL_ERR("Data file ftruncate failure, pos: "
		      "%" PRId64 ": %s", data_offset,
		      strerror(errno));
	    vtl_readonly = (errno == ENOSPC || errno == EROFS);
	    return -1;
	  }
	}

	/* Update the filemark map removing any filemarks which will be
	   overwritten.  Rewrite the filemark map so that the on-disk image
	   of the map is consistent with the new sizes of the other two files.
	*/

	for (i = 0; i < meta.filemark_count; i++) {
		if (filemarks == NULL) {
			MHVTL_DBG(2, "No filemarks array available!");
			return 0;
		}
	  
		MHVTL_DBG(2, "filemarks[%d] %d", i, filemarks[i]);
		if (filemarks[i] >= blk_number) {
			MHVTL_DBG(2, "Setting filemark_count from %d to %d",
					meta.filemark_count, i);
			meta.filemark_count_save = meta.filemark_count;
			meta.filemark_count = i;
			return rewrite_meta_file(1 /* log error */);
		}
	}

	return 0;
}

static int check_filemarks_alloc(uint32_t count)
{
	uint32_t new_size;

	/* See if we have enough space allocated to hold 'count' filemarks.
	   If not, realloc now.
	*/

	if (count > (uint32_t)filemark_alloc) {
		new_size = ((count + FM_DELTA - 1) / FM_DELTA) * FM_DELTA;

		filemarks = (uint32_t *)realloc(filemarks,
						new_size * sizeof(*filemarks));
		if (filemarks == NULL) {
			MHVTL_ERR("filemark map realloc failed, %s\n",
				strerror(errno));
			return -1;
		}
		filemark_alloc = new_size;
	}
	return 0;
}

static int add_filemark(uint32_t blk_number)
{
	/*
	 * See if we have enough space remaining to add the new filemark.  If
	 * not, realloc now.
	 */

	if (check_filemarks_alloc(meta.filemark_count + 1)) {
	  return -1;
	}

	/* Save */
	meta.filemark_count_save = meta.filemark_count;

	/* Record */
	filemarks[meta.filemark_count++] = blk_number;

	/* Now rewrite the meta_header structure and the filemark map. */
	return rewrite_meta_file(1 /* log error */);
}

/*
 * Return 0 -> Not loaded.
 *        1 -> Load OK
 *        2 -> format corrupt.
 */

int rewind_tape(uint8_t *sam_stat)
{
  	uint32_t blk_number;

	blk_number = raw_pos.hdr.blk_number;

	/* Sync point */
	MHVTL_DBG(3, "rewind_tape: flush: ibuf.idx %d, ibuf.data_written %ld, "
		  "blk_number %d", ibuf.idx, ibuf.data_written, blk_number);
	
	if (flush_index(sam_stat)) {
	  MHVTL_DBG(3, "rewind_tape: flush_index returned error, rofs.");
	  /* do not return error, as it is not a write operation */  
	}
	
	cartsync();

	if (!tape_loaded(sam_stat))
		return -1;

	if (read_header(0, sam_stat))
		return -1;

	switch (mam.MediumType) {
	case MEDIA_TYPE_CLEAN:
		OK_to_write = 0;
		break;
	case MEDIA_TYPE_WORM:
		/* Check if this header is a filemark and the next header
		 *  is End of Data. If it is, we are OK to write
		 */

		if (raw_pos.hdr.blk_type == B_EOD ||
		    (raw_pos.hdr.blk_type == B_FILEMARK && eod_blk_number == 1))
			OK_to_write = 1;
		else
			OK_to_write = 0;
		break;
	case MEDIA_TYPE_DATA:
		OK_to_write = 1;	/* Reset flag to OK. */
		break;
	}

	MHVTL_DBG(1, "Media is%s writable", (OK_to_write) ? "" : " not");

	return 1;
}

/*
 * Returns:
 * == 0, success
 * != 0, failure
*/

int position_to_eod(uint8_t *sam_stat)
{
	if (!tape_loaded(sam_stat))
		return -1;

	if (read_header(eod_blk_number, sam_stat))
		return -1;

	if (mam.MediumType == MEDIA_TYPE_WORM)
		OK_to_write = 1;

	return 0;
}

/*
 * Returns:
 * == 0, success
 * != 0, failure
*/

int position_to_block(uint32_t blk_number, uint8_t *sam_stat)
{
	if (!tape_loaded(sam_stat))
		return -1;

	MHVTL_DBG(2, "Position to block %d", blk_number);

	if (mam.MediumType == MEDIA_TYPE_WORM)
		OK_to_write = 0;

	if (blk_number > eod_blk_number) {
		sam_blank_check(E_END_OF_DATA, sam_stat);
		MHVTL_DBG(1, "End of data detected while positioning");
		return position_to_eod(sam_stat);
	}

	/* Treat a position to block zero specially, as it has different
	   semantics than other blocks when the tape is WORM.
	*/

	if (blk_number == 0)
		return rewind_tape(sam_stat);
	else
		return read_header(blk_number, sam_stat);
}

/*
 * Returns:
 * == 0, success
 * != 0, failure
*/

int position_blocks_forw(uint64_t count, uint8_t *sam_stat)
{
	uint32_t residual;
	uint32_t blk_target;
	unsigned int i;

	if (!tape_loaded(sam_stat))
		return -1;

	if (mam.MediumType == MEDIA_TYPE_WORM)
		OK_to_write = 0;

	blk_target = raw_pos.hdr.blk_number + count;

	/* Find the first filemark forward from our current position, if any. */

	for (i = 0; i < meta.filemark_count; i++) {
		MHVTL_DBG(3, "filemark at %ld", (unsigned long)filemarks[i]);
		if (filemarks[i] >= raw_pos.hdr.blk_number)
			break;
	}

	/* If there is one, see if it is between our current position and our
	   desired destination.
	*/

	if (i < meta.filemark_count) {
		if (filemarks[i] >= blk_target)
			return position_to_block(blk_target, sam_stat);

		residual = blk_target - raw_pos.hdr.blk_number + 1;
		if (read_header(filemarks[i] + 1, sam_stat))
			return -1;

		MHVTL_DBG(1, "Filemark encountered: block %d", filemarks[i]);
		sam_no_sense(SD_FILEMARK, E_MARK, sam_stat);
		put_unaligned_be32(residual, &sense[3]);
		return -1;
	}

	if (blk_target > eod_blk_number) {
		residual = blk_target - eod_blk_number;
		if (read_header(eod_blk_number, sam_stat))
			return -1;

		MHVTL_DBG(1, "EOD encountered");
		sam_blank_check(E_END_OF_DATA, sam_stat);
		put_unaligned_be32(residual, &sense[3]);
		return -1;
	}

	return position_to_block(blk_target, sam_stat);
}

/*
 * Returns:
 * == 0, success
 * != 0, failure
*/

int position_blocks_back(uint64_t count, uint8_t *sam_stat)
{
	uint32_t residual;
	uint32_t blk_target;
	int i = -1;
	unsigned int num_filemarks = meta.filemark_count;

	if (!tape_loaded(sam_stat))
		return -1;

	if (mam.MediumType == MEDIA_TYPE_WORM)
		OK_to_write = 0;

	MHVTL_DBG(2, "Position before movement: %d", raw_pos.hdr.blk_number);

	if (count < raw_pos.hdr.blk_number)
		blk_target = raw_pos.hdr.blk_number - count;
	else
		blk_target = 0;

	/* Find the first filemark prior to our current position, if any. */

	if (num_filemarks > 0) {
		for (i = num_filemarks - 1; i >= 0; i--) {
			MHVTL_DBG(3, "filemark at %ld",
						(unsigned long)filemarks[i]);
			if (filemarks[i] < raw_pos.hdr.blk_number)
				break;
		}
	}

	/* If there is one, see if it is between our current position and our
	   desired destination.
	*/
	if (i >= 0) {
		if (filemarks[i] < blk_target)
			return position_to_block(blk_target, sam_stat);

		residual = raw_pos.hdr.blk_number - blk_target;
		if (read_header(filemarks[i], sam_stat))
			return -1;

		MHVTL_DBG(2, "Filemark encountered: block %d", filemarks[i]);
		sam_no_sense(SD_FILEMARK, E_MARK, sam_stat);
		put_unaligned_be32(residual, &sense[3]);
		return -1;
	}

	if (count > raw_pos.hdr.blk_number) {
		residual = count - raw_pos.hdr.blk_number;
		if (read_header(0, sam_stat))
			return -1;

		MHVTL_DBG(1, "BOM encountered");
		sam_no_sense(SD_EOM, E_BOM, sam_stat);
		put_unaligned_be32(residual, &sense[3]);
		return -1;
	}

	return position_to_block(blk_target, sam_stat);
}

/*
 * Returns:
 * == 0, success
 * != 0, failure
*/

int position_filemarks_forw(uint64_t count, uint8_t *sam_stat)
{
	uint32_t residual;
	unsigned int i;

	if (!tape_loaded(sam_stat))
		return -1;

	if (mam.MediumType == MEDIA_TYPE_WORM)
		OK_to_write = 0;

	/* Find the block number of the first filemark greater than our
	   current position.
	*/

	for (i = 0; i < meta.filemark_count; i++)
		if (filemarks[i] >= raw_pos.hdr.blk_number)
			break;

	if (i + count - 1 < meta.filemark_count)
		return position_to_block(filemarks[i + count - 1] + 1, sam_stat);
	else {
		residual = i + count - meta.filemark_count;
		if (read_header(eod_blk_number, sam_stat))
			return -1;

		sam_blank_check(E_END_OF_DATA, sam_stat);
		put_unaligned_be32(residual, &sense[3]);
		return -1;
	}
}

/*
 * Returns 1 if cart is exported to cloud
 * else return 0
 */
int exported_to_cloud(void)
{
	if (exportfile >= 0) {
		// Cart exported to cloud
		return 1;
	}
	return 0;
}

/*
 * Returns:
 * == 0, success
 * != 0, failure
*/

int position_filemarks_back(uint64_t count, uint8_t *sam_stat)
{
	uint32_t residual;
	int i;

	if (!tape_loaded(sam_stat))
		return -1;

	if (mam.MediumType == MEDIA_TYPE_WORM)
		OK_to_write = 0;

	/* Find the block number of the first filemark less than our
	   current position.
	*/

	for (i = meta.filemark_count - 1; i >= 0; i--)
		if (filemarks[i] < raw_pos.hdr.blk_number)
			break;

	if (i + 1 >= count)
		return position_to_block(filemarks[i - count + 1], sam_stat);
	else {
		residual = count - i - 1;
		if (read_header(0, sam_stat))
			return -1;

		sam_no_sense(SD_EOM, E_BOM, sam_stat);
		put_unaligned_be32(residual, &sense[3]);
		return -1;
	}
}

/*
 * Writes data in struct MAM back to beginning of metafile..
 * Returns 0 if nothing written or -1 on error
 */

int rewriteMAM(uint8_t *sam_stat)
{
	loff_t nwrite = 0;

	if (!tape_loaded(sam_stat))
		return -1;

	if (vtl_readonly) { 
		MHVTL_ERR("WARNING: VTL is in read-only mode!");
	}

	if (exportfile >= 0) {
		// Cart exported to cloud, avoid MAM rewrites
		MHVTL_LOG("Avoiding MAM rewrite as cart is exported to cloud");
		return 0;
	}

	/* 
	 * Rewrite MAM data. Even if it is in read-only mode due
	 * to no space, this should succeed as it is an overwrite
	 */
	nwrite = pwrite(metafile, &mam, sizeof(mam), 0);
	if (nwrite != sizeof(mam)) {
		sam_medium_error(E_MEDIUM_FMT_CORRUPT, sam_stat);
		return -1;
	}

	return nwrite;
}

/*
 * Returns:
 * == 0, the new PCL was successfully created.
 * == 2, could not create the directory or some file(s)
 * == 1, an error occurred.
*/
int create_tape(const char *pcl, const struct MAM *mamp, uint8_t *sam_stat,
		const char *home_dir, int *out_err)
{
	struct stat data_stat;
	char *newMedia = NULL;
	char newMedia_data[1024];
	char newMedia_indx[1024];
	char newMedia_meta[1024];
	int rc = 0;
	struct passwd *pw;
	
	/* Attempt to create the new PCL.  This will fail if the PCL's directory
	   or any of the PCL's three files already exist, leaving any existing
	   files as they were.
	*/

	pw = getpwnam(USR);
	if (!pw) {
	  *out_err = errno;
	  MHVTL_ERR("Failed to get UID for user '%s': %s", USR,
		    strerror(errno));
	  return 1;
	}
	
	if (asprintf(&newMedia, "%s/%s", home_dir, pcl) < 0) {
		perror("Could not allocate memory");
		exit(1);
	}

	snprintf(newMedia_data, ARRAY_SIZE(newMedia_data), "%s/data", newMedia);
	snprintf(newMedia_indx, ARRAY_SIZE(newMedia_indx), "%s/indx", newMedia);
	snprintf(newMedia_meta, ARRAY_SIZE(newMedia_meta), "%s/meta", newMedia);

	/* Check if data file already exists, nothing to create */
	if (stat(newMedia_data, &data_stat) != -1) {
		if (verbose)
		  printf("Warning: Data file already exists for new media\n");
		goto free_strings;
	}

	if (verbose)
		printf("Creating new media directory: %s\n", newMedia);

	umask(0007);
	
	rc = mkdir(newMedia, S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP|S_ISGID);
	if (rc) {
		if (errno != EEXIST) {
		  *out_err = errno;
		  MHVTL_ERR("Failed to create directory %s: %s",
			    newMedia, strerror(errno));
		  rc = 2;
		  goto free_strings;
		}
		/* No need to fail just because the parent dir exists */
		rc = 0;
	}

	if (verbose)
		printf("Creating new media data file: %s\n", newMedia_data);

	/*
	 * Don't really care if chown() fails or not..
	 * But lets try anyway
	 */
	chown(newMedia, pw->pw_uid, pw->pw_gid);
	
	datafile = creat(newMedia_data, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
	if (datafile == -1) {
	  *out_err = errno;
	  MHVTL_ERR("Failed to create file %s: %s", newMedia_data,
		    strerror(errno));
	  rmdir(newMedia);
	  rc = 2;
	  goto free_strings;
	}

	if (verbose)
		printf("Creating new media index file: %s\n", newMedia_indx);
	indxfile = creat(newMedia_indx, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
	if (indxfile == -1) {
	  *out_err = errno;
	  MHVTL_ERR("Failed to create file %s: %s", newMedia_indx,
		    strerror(errno));
	  unlink(newMedia_data);
	  rmdir(newMedia);
	  rc = 2;
	  goto cleanup;
	}
	
	if (verbose)
		printf("Creating new media meta file: %s\n", newMedia_meta);
	
	metafile = creat(newMedia_meta, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
	if (metafile == -1) {
	  *out_err = errno;
	  MHVTL_ERR("Failed to create file %s: %s", newMedia_meta,
		    strerror(errno));
	  unlink(newMedia_data);
	  unlink(newMedia_indx);
	  rmdir(newMedia);
	  rc = 2;
	  goto cleanup;
	}

	chown(newMedia_data, pw->pw_uid, pw->pw_gid);
	chown(newMedia_indx, pw->pw_uid, pw->pw_gid);
	chown(newMedia_meta, pw->pw_uid, pw->pw_gid);

	MHVTL_DBG(1, "%s files created", newMedia);

	/* Write the meta file consisting of the MAM and the meta_header
	   structure with the filemark count initialized to zero.
	*/
	mam = *mamp;

	memset(&meta, 0, sizeof(meta));
	meta.filemark_count = 0;

	if (write(metafile, &mam, sizeof(mam)) != sizeof(mam) ||
		    write(metafile, &meta, sizeof(meta)) != sizeof(meta)) {
	  *out_err = errno;
	  MHVTL_ERR("Failed to initialize file %s: %s", newMedia_meta,
		    strerror(errno));
	  unlink(newMedia_data);
	  unlink(newMedia_indx);
	  unlink(newMedia_meta);
	  rmdir(newMedia);
	  rc = 1;
	}

cleanup:
	if (datafile >= 0) {
	  fsync(datafile);
	  close(datafile);
	  datafile = -1;
	}
	if (indxfile >= 0) {
	  fsync(indxfile);
	  close(indxfile);
	  indxfile = -1;
	}
	if (metafile >= 0) {
	  fsync(metafile);
	  close(metafile);
	  metafile = -1;
	}

free_strings:
	if (newMedia)
	  free(newMedia);
	
	return rc;
}

static int __vtlck_metafile()
{
  uint64_t exp_size;
  int rc = 0;
  struct stat meta_stat;
  
  /* Now recompute the correct size of the meta file. */
  exp_size = sizeof(mam) + sizeof(meta) +
    (meta.filemark_count * sizeof(*filemarks));
  
  if (fstat(metafile, &meta_stat) < 0) {
    MHVTL_ERR("stat of pcl meta file %s/meta failed: %s\n",
	      currentPCL, strerror(errno));
    rc = 3;
    goto failed;
  }


  MHVTL_LOG("VTLCK: pcl meta file %s/meta: filemark_count: %u "
	    "filemark_count_save: %u "
	    "exp_size[sizeof(mam) + sizeof(meta) + "
	    "(meta.filemark_count * sizeof(*filemarks))]:%"PRId64 " "
	    "meta_stat.st_size:%"PRId64 "\n",
	    currentPCL, meta.filemark_count,
	    meta.filemark_count_save, exp_size,
	    meta_stat.st_size);
  
  if ((uint32_t)meta_stat.st_size != exp_size) {
    if ((uint32_t)meta_stat.st_size > exp_size) {
      MHVTL_LOG("VTLCK: FM: pcl meta file %s/meta is not the correct length, "
		"expected %" PRId64 ", actual %" PRId64
		" auto-correcting ...\n", 
		currentPCL, exp_size, meta_stat.st_size);
      if (ftruncate(metafile, exp_size) < 0) {
	if (errno == EROFS) {
	  MHVTL_LOG("Ignoring the ftruncate error because VTL is RO.\n");
	} else {
	  MHVTL_ERR("Error truncating data: %s\n", strerror(errno));
	  rc = 2;
	  goto failed;
	}
      }
    } else {
      MHVTL_ERR("pcl meta file %s/meta is not the correct length, "
		"expected %" PRId64 ", actual %" PRId64"\n",
		currentPCL, exp_size, meta_stat.st_size);
      rc = 2;
      goto failed;
    }
  }
  
 failed:
  return rc;
}

static int __vtlck_meta_indx(uint8_t *sam_stat)
{
  int rc = 0;

  /* Make sure that the filemark map is consistent with the size of the
     indx file.
  */
  if (meta.filemark_count &&
      filemarks[meta.filemark_count - 1] + 1 != eod_blk_number ) {
    MHVTL_LOG("VTLCK: pcl indx file %s/indx has improper length"
	      " as compared to the meta file - auto-correcting to"
	      " last file mark position!"
	      " meta.filemark_count: %u "
	      " filemarks[meta.filemark_count - 1] + 1: %u "
	      " eod_blk_number: %u\n",
	      currentPCL, meta.filemark_count,
	      filemarks[meta.filemark_count - 1] + 1,
	      eod_blk_number);
    if (eod_blk_number) {
      if (eod_blk_number > (filemarks[meta.filemark_count - 1])) {
	eod_blk_number = filemarks[meta.filemark_count - 1] + 1;
	MHVTL_LOG("EOD %u -> last file mark position %u\n",  
		  eod_blk_number,
		  filemarks[meta.filemark_count - 1]);
      } else {
	/*
	 * Try going back to the last sane FM
	 */
	while (meta.filemark_count >= 2) {
	  meta.filemark_count--;
	  if (eod_blk_number < filemarks[meta.filemark_count - 1]){
	    continue;
	  } else if (eod_blk_number > filemarks[meta.filemark_count - 1]) {
	    MHVTL_LOG("EOD %u is > last file mark position %u\n",
		      eod_blk_number,
		      filemarks[meta.filemark_count - 1]);
	    meta.filemark_count_save = meta.filemark_count;
	    rewrite_meta_file(0);
	    /* clear the readonly flag as this is not a tape_write op */ 
	    vtl_readonly = 0; 
	    if ((rc = __vtlck_metafile()) > 0){
	      goto failed;
	    }
	    
	    break;
	  }
	}
      }
    }
  }
  /* Read in the last raw_header struct from the indx file and use that
   * to validate the correct size of the data file.
   */
  
  if (eod_blk_number == 0) {
    eod_data_offset = 0;
    meta.filemark_count = 0;
    meta.filemark_count_save = 0;
  } else {
    if (read_header(eod_blk_number - 1, sam_stat)) {
      rc = 3;
      goto failed;
    }
    eod_data_offset = raw_pos.data_offset +
      raw_pos.hdr.disk_blk_size;
  }
  
 failed:
  return rc;
}

/*
 * Attempt to load PCL - i.e. Open datafile and read in BOT header & MAM
 *
 * Returns:
 * == 0 -> Load OK
 * == 1 -> Another tape already loaded.
 * == 2 -> format corrupt.
 * == 3 -> cartridge does not exist or cannot be opened.
 */

int load_tape(const char *pcl, uint8_t *sam_stat, const char *home_dir)
{
	char pcl_data[1024]= { 0 }, pcl_indx[1024]= { 0 },
	  pcl_meta[1024], pcl_export[1024] = { 0 };
	struct stat data_stat, indx_stat, meta_stat;
	uint64_t exp_size;
	size_t	io_size;
	loff_t nread;
	int rc = 0;
	int null_media_type;
	char touch_file[128];
	int ret = 0;
	struct passwd *pw;

	/* KFRDEBUG - sam_stat needs updates in lots of places here. */

	/* If some other PCL is already open, return. */
	if (datafile >= 0)
		return 1;

	/* Open all three files and stat them to get their current sizes. */
	sprintf(currentbar, pcl);
	snprintf(currentPCL, ARRAY_SIZE(currentPCL), "%s/%s",
		 home_dir, pcl);
	
	snprintf(pcl_data, ARRAY_SIZE(pcl_data), "%s/data", currentPCL);
	snprintf(pcl_indx, ARRAY_SIZE(pcl_indx), "%s/indx", currentPCL);
	snprintf(pcl_meta, ARRAY_SIZE(pcl_meta), "%s/meta", currentPCL);
	strncpy(pcl_export, home_dir, (strlen(home_dir) - 6));
	strncat(pcl_export, "/pool/", strlen("/pool/"));
	strncat(pcl_export, pcl, strlen(pcl));
	
	MHVTL_DBG(2, "Opening media: %s\n", pcl);
	MHVTL_LOG("Opening media: [%s], pcl_data: [%s]\n", pcl, pcl_data);
	datafile = open(pcl_data, O_RDWR|O_LARGEFILE);
	if (datafile == -1) {
	  /* 
	   * Could be a result of something that went wrong previously
	   * Recreate the data file at this time.
	   */
	  MHVTL_ERR("data file %s is not present, recreate it !: %s\n", pcl_data,
		    strerror(errno));
	  
	  MHVTL_DBG(3, "data file %s is not present, recreate it !\n", pcl_data);		
	  datafile = creat(pcl_data, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP); 
	  if (datafile < 0 &&  errno != EEXIST){
	    MHVTL_ERR("unable to create %s: %s", pcl_data, strerror(errno));
	    vtl_readonly = (errno == ENOSPC || errno == EROFS);
	    rc = 3;
	    goto failed;
	  }
	  close(datafile);
	  pw = getpwnam(USR);	/* Find UID for user 'vtl' */
	  if (!pw) {
	    MHVTL_ERR("Failed to get UID for user '%s': %s\n", USR,
		      strerror(errno));
	    rc = 3;
	    goto failed;
	  }
	  chown(pcl_data, pw->pw_uid, pw->pw_gid);
	  datafile = open(pcl_data, O_RDWR|O_LARGEFILE); 
	  if (datafile < 0){
	    MHVTL_ERR("open of pcl %s file %s failed, %s\n", pcl,
		      pcl_data, strerror(errno));
	    rc = 3;
	    goto failed;
	  }
	}
	
	indxfile = open(pcl_indx, O_RDWR|O_LARGEFILE);
	if (indxfile == -1) {
		MHVTL_ERR("open of pcl %s file %s failed, %s\n", pcl,
			pcl_indx, strerror(errno));
		rc = 3;
		goto failed;
	}
	metafile = open(pcl_meta, O_RDWR|O_LARGEFILE);
	if (metafile == -1) {
		MHVTL_ERR("open of pcl %s file %s failed, %s\n", pcl,
			pcl_meta, strerror(errno));
		rc = 3;
		goto failed;
	}

	exportfile = open(pcl_export, O_RDONLY);
	if (exportfile == -1) {
		// Not an error if export file doesn't exist.
		MHVTL_LOG("open of export marker file %s for pcl %s file failed, %s\n",
			  pcl_export, pcl, strerror(errno));
	}
	
	if (fstat(datafile, &data_stat) < 0) {
		MHVTL_ERR("stat of pcl %s file %s failed: %s\n", pcl,
			pcl_data, strerror(errno));
		rc = 3;
		goto failed;
	}

	if (fstat(indxfile, &indx_stat) < 0) {
		MHVTL_ERR("stat of pcl %s file %s failed: %s\n", pcl,
			pcl_indx, strerror(errno));
		rc = 3;
		goto failed;
	}

	if (fstat(metafile, &meta_stat) < 0) {
		MHVTL_ERR("stat of pcl %s file %s failed: %s\n", pcl,
			pcl_meta, strerror(errno));
		rc = 3;
		goto failed;
	}

	ibuf.offset = ibuf.idx = ibuf.data_written = 0;
	memset(ibuf.buf, 0, INDEX_BUF_SIZE);
	
	/* Verify that the metafile size is at least reasonable. */

	exp_size = sizeof(mam) + sizeof(meta);
	MHVTL_LOG("Media[%s]: exp_size[sizeof(mam) + sizeof(meta)]:%"PRId64" "
		  "file [%s] size[meta_stat.st_size]:%"PRId64"\n",
		  pcl, exp_size, pcl_meta, meta_stat.st_size);
	if ((uint32_t)meta_stat.st_size < exp_size) {
		MHVTL_ERR("sizeof(mam) + sizeof(meta) - "
			  "pcl %s file %s is not the correct length, "
			  "expected at least %" PRId64 ", actual %" PRId64
			  "\n",
			  currentPCL, pcl_meta, exp_size, meta_stat.st_size);
		rc = 2;
		goto failed;
	}

	/* Read in the MAM and sanity-check it. */
	nread = read(metafile, &mam, sizeof(mam));
	if (nread < 0) {
		MHVTL_ERR("Error reading pcl %s MAM from metafile: %s\n",
			pcl, strerror(errno));
		rc = 2;
		goto failed;
	} else if (nread != sizeof(mam)) {
		MHVTL_ERR("Error reading pcl %s MAM from metafile: "
			"unexpected read length\n", pcl);
		rc = 2;
		goto failed;
	}

	null_media_type = (mam.MediumType == MEDIA_TYPE_NULL) ? 1 : 0;

	if (mam.tape_fmt_version != TAPE_FMT_VERSION) {
		MHVTL_ERR("pcl %s MAM contains incorrect media format\n", pcl);
		sam_medium_error(E_MEDIUM_FMT_CORRUPT, sam_stat);
		rc = 2;
		goto failed;
	}

	/* Read in the meta_header structure and sanity-check it. */

	nread = read(metafile, &meta, sizeof(meta));
	if (nread < 0) {
		MHVTL_ERR("Error reading pcl %s meta_header from "
			"metafile: %s\n", pcl, strerror(errno));
		rc = 2;
		goto failed;
	} else if (nread != sizeof(meta)) {
		MHVTL_ERR("Error reading pcl %s meta header from "
			"metafile: unexpected read length\n", pcl);
		rc = 2;
		goto failed;
	}

	if ((rc =__vtlck_metafile()) > 0) {
	  goto failed;
	}

	/* See if we have allocated enough space for the actual number of
	   filemarks on the tape.  If not, realloc now.
	*/

	if (check_filemarks_alloc(meta.filemark_count)) {
	  rc = 3;
	  goto failed;
	}

	/* Now read in the filemark map. */

	io_size = meta.filemark_count * sizeof(*filemarks);
	if (io_size) {
		nread = read(metafile, filemarks, io_size);
		if (nread < 0) {
			MHVTL_ERR("Error reading pcl %s filemark map from "
				"metafile: %s\n", pcl, strerror(errno));
			rc = 2;
			goto failed;
		} else if ((size_t)nread != io_size) {
			MHVTL_ERR("Error reading pcl %s filemark map from "
				"metafile: unexpected read length\n", pcl);
			rc = 2;
			goto failed;
		}
	}

	/* Use the size of the indx file to work out where the virtual
	   B_EOD block resides.
	*/

	if ((indx_stat.st_size % sizeof(struct raw_header)) != 0) {
		MHVTL_ERR("pcl %s indx file has improper length, indicating "
			"possible file corruption\n", pcl);
		rc = 2;
		goto failed;
	}
	eod_blk_number = indx_stat.st_size / sizeof(struct raw_header);

	MHVTL_LOG("Media[%s]: eod_blk_number:%u"
		  " indx_stat.st_size:%"PRId64" sizeof(struct raw_header):%"PRId64
		  "\n", pcl, eod_blk_number, indx_stat.st_size, sizeof(struct raw_header));
	
	if ((rc =__vtlck_meta_indx(sam_stat)) > 0) {
	  goto failed;
	}

	if (null_media_type) {
	  MHVTL_LOG("Loaded NULL media type\n");	/* Skip check */
	} else {
	  /*
	   * If the EOD last recorded is less than the data file,
	   * then adjust back the data file.
	   */
	  if ((uint64_t)data_stat.st_size > eod_data_offset) {
	    MHVTL_LOG("VTLCK: pcl %s file %s is not the correct length, "
		      "expected %" PRId64 ", actual %" PRId64"\n", pcl,
		      pcl_data, eod_data_offset, data_stat.st_size);
	    MHVTL_LOG("Truncating the datafile %s to %" PRId64"\n", 
		      pcl_data, eod_data_offset);
	    if (ftruncate(datafile, eod_data_offset) < 0) {
	      if (errno == EROFS) {
		MHVTL_LOG("Ignoring the ftruncate error because VTL is RO.\n");
	      } else {
		MHVTL_ERR("Error truncating data: %s\n", strerror(errno));
		rc = 2;
		goto failed;
	      }
	    }
	  } else if ((uint64_t)data_stat.st_size < eod_data_offset) {
	    MHVTL_LOG("VTLCK: pcl %s file %s is not the correct length, "
		      "expected %" PRId64 ", actual %" PRId64"\n", pcl,
		      pcl_data, eod_data_offset, data_stat.st_size);
	    while (eod_blk_number > 1 && eod_data_offset != (uint64_t)data_stat.st_size) {
	      eod_blk_number--;
	      if (read_header(eod_blk_number - 1, sam_stat)) {
		rc = 3;
		goto failed;
	      }
	      eod_data_offset = raw_pos.data_offset +
		raw_pos.hdr.disk_blk_size;
	      if (eod_data_offset > (uint64_t)data_stat.st_size) {
		continue;
	      } else if (eod_data_offset <= (uint64_t)data_stat.st_size) {
		MHVTL_LOG("Truncating the datafile %s to %" PRId64"\n",
			  pcl_data, eod_data_offset);
		if (ftruncate(datafile, eod_data_offset) < 0) {
		  if (errno == EROFS) {
		    MHVTL_ERR("Ignoring the ftruncate error because VTL is RO.\n");
		  } else {
		    MHVTL_ERR("Error truncating data: %s\n", strerror(errno));
		    rc = 2;
		    goto failed;
		  }
		}
		if ((rc = __vtlck_meta_indx(sam_stat) > 0)){
		  goto failed;
		}
		break;
	      }
	    }
	  }
	}
	
	/* 
	 * Give a hint to the kernel that data, once written, tends not to be
	 * accessed again immediately.
	 */

	posix_fadvise(indxfile, 0, 0, POSIX_FADV_DONTNEED);
	posix_fadvise(datafile, 0, 0, POSIX_FADV_DONTNEED);

	/* Initialise SAM STATUS */
	*sam_stat = SAM_STAT_GOOD;

	/* Now initialize raw_pos by reading in the first header, if any. */
	if (read_header(0, sam_stat)) {
		rc = 3;
		goto failed;
	}

	return 0;

failed:
	if (datafile >= 0) {
		close(datafile);
		datafile = -1;
	}
	if (indxfile >= 0) {
		close(indxfile);
		indxfile = -1;
	}
	if (metafile >= 0) {
		close(metafile);
		metafile = -1;
	}
	if (exportfile >= 0) {
		close(exportfile);
		exportfile = -1;
	}
	return rc;
}

void zero_filemark_count(void)
{
	if (filemarks)
	  free(filemarks);
	filemark_alloc = 0;
	filemarks = NULL;

	/* Save */
	meta.filemark_count_save = meta.filemark_count;
	meta.filemark_count = 0;
	rewrite_meta_file(1 /* log error */);
}

int format_tape(uint8_t *sam_stat)
{
	uint32_t blk_number;

	blk_number = raw_pos.hdr.blk_number;
	if (!tape_loaded(sam_stat))
		return -1;
	MHVTL_DBG(3, "format_tape: flush: ibuf.idx %d, ibuf.data_written %ld, "
		  "blk_number %d", ibuf.idx, ibuf.data_written, blk_number);
	if (flush_index(sam_stat)) {
	  return -1;
	}
	
	if (check_for_overwrite(sam_stat))
		return -1;

	zero_filemark_count();

	cartsync();

	ibuf.offset = raw_pos.hdr.blk_number * sizeof(raw_pos);
	
	return mkEODHeader(raw_pos.hdr.blk_number, raw_pos.data_offset);
}

/*
 * Returns:
 * == 0, success
 * != 0, failure
*/

int write_filemarks(uint32_t count, uint8_t *sam_stat)
{
	uint32_t blk_number;
	uint64_t data_offset;
	ssize_t nwrite;

	if (vtl_readonly) {
		sam_data_protect(E_WRITE_PROTECT, sam_stat);
		MHVTL_ERR("Cannot write file marks in read-only mode!");
		return -1;
	}
	
	if (!tape_loaded(sam_stat)) {
		return -1;
	}
	
	/* Applications assume that writing a filemark (even writing zero
	   filemarks) will force-flush any data buffered in the drive to media
	   so that after the write-filemarks call returns there is no
	   possibility that any data previously written could be lost due
	   to a power hit.  Provide a similar guarantee here.
	*/

	if (count == 0) {
		MHVTL_DBG(2, "Flushing data - 0 filemarks written");
		fsync(datafile);
		fsync(indxfile);
		fsync(metafile);

		return 0;
	}

	if (check_for_overwrite(sam_stat))
		return -1;

	/* Preserve existing raw_pos data we need, then clear raw_pos and
	   fill it in with new data.
	*/

	blk_number = raw_pos.hdr.blk_number;
	data_offset = raw_pos.data_offset;

	memset(&raw_pos, 0, sizeof(raw_pos));

	raw_pos.data_offset = data_offset;

	raw_pos.hdr.blk_type = B_FILEMARK;	/* Header type */
	raw_pos.hdr.blk_flags = 0;
	raw_pos.hdr.blk_number = blk_number;
	raw_pos.hdr.blk_size = 0;
	raw_pos.hdr.disk_blk_size = 0;

	/* Flush out the existing index buffer.*/
	MHVTL_DBG(3, "FM: flush: ibuf.idx %d, ibuf.data_written %ld, blk_number %d, "
		  "count %d", ibuf.idx, ibuf.data_written, blk_number, count);
	if (flush_index(sam_stat)) {
	  return -1;
	}
	
	/* Now write out one header per filemark. */

	for ( ; count > 0; count--, blk_number++) {
		raw_pos.hdr.blk_number = blk_number;

		MHVTL_DBG(3, "Writing filemark: block %d", blk_number);

		nwrite = pwrite(indxfile, &raw_pos, sizeof(raw_pos),
			blk_number * sizeof(raw_pos));
		if (nwrite != sizeof(raw_pos)) {
			sam_medium_error(E_WRITE_ERROR, sam_stat);
			MHVTL_ERR("Index file write failure,"
					" pos: %" PRId64 ": %s",
				(uint64_t)blk_number * sizeof(raw_pos),
				strerror(errno));
			vtl_readonly = (errno == ENOSPC || errno == EROFS);
			return -1;
		}
		add_filemark(blk_number);
		ibuf.offset += sizeof(raw_pos);
	}

	/* 
	 * Provide a force-flush guarantee and before
	 * going back to the initiator, make sure we
	 * tried our best.
	 */
	fsync(datafile);
	fsync(indxfile);
	fsync(metafile);

	return mkEODHeader(blk_number, data_offset);
}

/* mode : 1 readonly
   0 readwrite */
void set_drive_iomode(int mode)
{
  if (mode != vtl_readonly) {
    MHVTL_LOG("Setting drive io-mode to %s", (mode) ? "readonly" : "readwrite");
    vtl_readonly = mode;
  }
}

int is_read_only()
{
  return vtl_readonly;
}

int write_tape_block(const uint8_t *buffer, uint32_t blk_size,
		     uint32_t comp_size, const struct encryption *encryptp,
		     uint8_t comp_type, uint8_t null_media_type,
		     uint32_t crc, uint8_t *sam_stat)
{
  	uint32_t max_blk_number;
	uint32_t blk_number, disk_blk_size, tmp_blk_size;
	uint64_t data_offset, tmp_offset;
	ssize_t nwrite = 0;
	uint8_t *tmp_buf;
	int idx_hdr_size = sizeof(raw_pos);
	ssize_t n, bc = 0;
	
	/*
 	 * If the tape was loaded, this is what we should 
	 * return. If we loaded after tape drive went into
	 * read-only mode, then we set it in restrictions.
	 */
	if (vtl_readonly) {
		sam_data_protect(E_WRITE_PROTECT, sam_stat);
		MHVTL_ERR("Cannot write data in read-only mode!");
		return -1;
	}

	/* Medium format limits to unsigned 32bit blks */
	max_blk_number = 0xfffffff0;

	if (!tape_loaded(sam_stat)) {
		return -1;
	}

	if (check_for_overwrite(sam_stat)) {
		return -1;
	}

	/* 
	 * Preserve existing raw_pos data we need, then clear out 
	 * raw_pos and fill it in with new data.
	 */

	blk_number = raw_pos.hdr.blk_number;
	data_offset = raw_pos.data_offset;

	if (blk_number > max_blk_number) {
		MHVTL_ERR("Too many tape blocks - 32bit overflow");
		return -1;
	}

	memset(&raw_pos, 0, sizeof(raw_pos));

	raw_pos.data_offset = data_offset;

	raw_pos.hdr.blk_type = B_DATA;	/* Header type */
	raw_pos.hdr.blk_flags = 0;
	raw_pos.hdr.blk_number = blk_number;
	raw_pos.hdr.blk_size = blk_size; /* Size of uncompressed data */

	raw_pos.hdr.uncomp_crc = crc;
	raw_pos.hdr.blk_flags |= BLKHDR_FLG_UNCOMPRESSED_CRC;
	
	MHVTL_DBG(2, "CRC is 0x%08x", crc);

	if (comp_size) {
		if (comp_type == LZO)
			raw_pos.hdr.blk_flags |= BLKHDR_FLG_LZO_COMPRESSED;
		else
			raw_pos.hdr.blk_flags |= BLKHDR_FLG_ZLIB_COMPRESSED;
		raw_pos.hdr.disk_blk_size = disk_blk_size = comp_size;
	} else
		raw_pos.hdr.disk_blk_size = disk_blk_size = blk_size;

	if (encryptp != NULL) {
		unsigned int i;

		raw_pos.hdr.blk_flags |= BLKHDR_FLG_ENCRYPTED;
		raw_pos.hdr.blk_encryption_info.ukad_length = encryptp->ukad_length;
		for (i = 0; i < encryptp->ukad_length; ++i)
			raw_pos.hdr.blk_encryption_info.ukad[i] = encryptp->ukad[i];

		raw_pos.hdr.blk_encryption_info.akad_length = encryptp->akad_length;
		for (i = 0; i < encryptp->akad_length; ++i)
			raw_pos.hdr.blk_encryption_info.akad[i] = encryptp->akad[i];

		raw_pos.hdr.blk_encryption_info.key_length = encryptp->key_length;
		for (i = 0; i < encryptp->key_length; ++i)
			raw_pos.hdr.blk_encryption_info.key[i] = encryptp->key[i];
	}

	if (null_media_type) {
		nwrite = disk_blk_size;
	} else {
	  /* Now write out both the data and the header. */
	  tmp_blk_size = disk_blk_size;
	  tmp_buf = (uint8_t *)buffer;
	  tmp_offset = data_offset;
	  
	  /* Limitation of underlying layer - has a max io size */
	  while (tmp_blk_size) {
	    bc = min((ssize_t)tmp_blk_size, (ssize_t)MAX_IOSZ);
	    
	    n = pwrite(datafile, tmp_buf, bc, tmp_offset);
	    if (n != bc) {
	      sam_medium_error(E_WRITE_ERROR, sam_stat);
	      MHVTL_ERR("Data file write failure, pos: %" PRId64 ": %s",
			data_offset, 
			strerror(errno));
	      vtl_readonly = (errno == ENOSPC);
	      /* Revert back by truncating and setting EOD */
	      MHVTL_ERR("Truncating data file: %"PRId64,
			data_offset);
	      if (ftruncate(datafile, data_offset) < 0) {
		MHVTL_ERR("Error truncating data: %s", strerror(errno));
	      }
	      mkEODHeader(blk_number, data_offset);
	      return -1;
	    }
	    
	    nwrite += n;
	    tmp_offset += bc;
	    tmp_buf += bc;
	    tmp_blk_size -= bc;
	  }
	  
	  assert(nwrite == disk_blk_size);
	  
	  memcpy(&ibuf.buf[ibuf.idx * idx_hdr_size], &raw_pos, idx_hdr_size);
	  ibuf.idx++;
	  ibuf.data_written += bc;
	  if (ibuf.idx == ((INDEX_BUF_SIZE)/idx_hdr_size - 1)) {
	    MHVTL_DBG(3, "Flush: ibuf.idx %d, ibuf.data_written %ld, "
		      "blk_number %d", ibuf.idx, ibuf.data_written, blk_number);
	    if (flush_index(sam_stat)) {
	      return -1;
	    }
	  }
	}

	MHVTL_DBG(3, "Successfully wrote block: %u", blk_number);

	return mkEODHeader(blk_number + 1, data_offset + disk_blk_size);
}

void unload_tape(uint8_t *sam_stat)
{
        uint32_t blk_number;
	blk_number = raw_pos.hdr.blk_number;

	if (datafile >= 0) {
		close(datafile);
		datafile = -1;
	}
	if (indxfile >= 0) {
	  MHVTL_DBG(3, "unload_tape: flush: ibuf.idx %d, ibuf.data_written %ld, "
		    "blk_number %d", ibuf.idx, ibuf.data_written, blk_number);
	  flush_index(sam_stat);
	  close(indxfile);
	  indxfile = -1;
	  ibuf.offset = ibuf.idx = ibuf.data_written = 0;
	  memset(ibuf.buf, 0, INDEX_BUF_SIZE);
	}
	if (metafile >= 0) {
	  rewrite_meta_file(0 /* do not log error */);
	  close(metafile);
	  metafile = -1;
	}
	if (exportfile >= 0) {
	  close(exportfile);
	  exportfile = -1;
	}
	if (filemarks)
	  free(filemarks);
	filemarks = NULL;
	filemark_alloc = 0;
	/* clear the readonly flag when tape is unloaded */
	vtl_readonly = 0;
}

uint32_t read_tape_block(uint8_t *buf, uint32_t buf_size, uint8_t *sam_stat)
{
  	loff_t nread, tmp_offset;
	uint32_t iosize, tmp_blk_size;
	uint8_t *tmp_buf;
	nread = 0;

	if (!tape_loaded(sam_stat))
		return -1;

	MHVTL_DBG(3, "Reading blk %ld, size: %d",
			(unsigned long)raw_pos.hdr.blk_number, buf_size);

	/* The caller should have already verified that this is a
	 * B_DATA block before issuing this read, so we shouldn't have to
	 * worry about B_EOD or B_FILEMARK here.
	 */
	if (raw_pos.hdr.blk_type == B_EOD) {
		sam_blank_check(E_END_OF_DATA, sam_stat);
		MHVTL_ERR("End of data detected while reading");
		return -1;
	}

	iosize = raw_pos.hdr.disk_blk_size;
	if (iosize > buf_size)
		iosize = buf_size;
	tmp_blk_size = iosize;
	tmp_buf = (uint8_t *)buf;
	tmp_offset = raw_pos.data_offset;

	while (tmp_blk_size) {
		ssize_t n, bc = min((ssize_t)tmp_blk_size, (ssize_t)MAX_IOSZ);

		n = pread(datafile, tmp_buf, bc, tmp_offset);
		if (n != bc) {
			MHVTL_ERR("Failed to read %d bytes (ret: %d): %s", 
				(int)bc, (int)n, strerror(errno));
			return -1;
		}
		nread += n;
		tmp_offset += bc;
		tmp_buf += bc;
		tmp_blk_size -= bc;
	}

	assert(nread == iosize);

	/* Now position to the following block. */
	MHVTL_DBG(3, "Read block, now positioning to next header: %d",
				raw_pos.hdr.blk_number + 1);
	if (read_header(raw_pos.hdr.blk_number + 1, sam_stat)) {
		MHVTL_ERR("Failed to read block header %d",
				raw_pos.hdr.blk_number + 1);
		return -1;
	}

	return nread;
}

uint64_t current_tape_offset(void)
{
	if (datafile != -1)
		return raw_pos.data_offset;

	return 0;
}

uint64_t current_tape_block(void)
{
	if (datafile != -1)
		return (uint64_t)c_pos->blk_number;
	return 0;
}

uint64_t filemark_count(void)
{
	return (uint64_t)meta.filemark_count;
}

/* static void enc_key_to_string(char *dst, uint8_t *key, int len) */
/* { */
/* 	char b[16]; */
/* 	int i; */

/* 	dst[0] = '\0'; */

/* 	for (i = 0; i < len; i++) { */
/* 		sprintf(b, "%02x", key[i]); */
/* 		strncat(dst, b, 16); */
/* 	} */
/* 	key[i] = '\0'; */
/* } */

void print_format(int *count, int *begin, int *end, unsigned int *size)
{
	if (*count) {
		if (*count > 1) {
			if (size)
				printf("%d-%d: %d data | ", *begin, *end, *size);
			else
				printf("%d-%d: FM | ", *begin, *end);
		} else {
			if (size)
				printf("%d: %d data | ", *begin, *size);
			else
				printf("%d: FM | ", *begin);
		}
		*count = *begin = *end = 0;
		if (size) *size = 0;
	}
}

void print_tape_layout(uint8_t *sam_stat)
{
	int dbcount = 0, fmcount = 0;
	int dbbegin = 0, dbend = 0;
	unsigned int size = 0;
	int fmbegin = 0, fmend = 0;
	int bno = 0;

	printf("\nLayout:\n");
	printf("| BOT | ");
	while (c_pos->blk_type != B_EOD) {
		bno++;
		switch (raw_pos.hdr.blk_type) {
		case B_DATA:
			print_format(&fmcount, &fmbegin, &fmend, NULL);
			if (size && (size != raw_pos.hdr.blk_size)) {
				print_format(&dbcount, &dbbegin, &dbend, &size);
			}
			if (!dbcount) {
				dbbegin = raw_pos.hdr.blk_number;
				size = raw_pos.hdr.blk_size;
				dbend = 0;
			} else {
				dbend = raw_pos.hdr.blk_number;
			}
			dbcount++;
			break;
		case B_FILEMARK:
			print_format(&dbcount, &dbbegin, &dbend, &size);
			if (!fmcount) {
				fmbegin = raw_pos.hdr.blk_number;
				fmend = 0;
			} else {
				fmend = raw_pos.hdr.blk_number;
			}
			fmcount++;
			break;
		case B_EOD:
			print_format(&dbcount, &dbbegin, &dbend, &size);
			print_format(&fmcount, &fmbegin, &fmend, NULL);
			printf("EOD |\n\n");
			break;
		case B_NOOP:
			//The while loop will stop before we reach here.
			break;
		default:
			break;
		}
		position_blocks_forw(1, sam_stat);
	}
	print_format(&dbcount, &dbbegin, &dbend, &size);
	print_format(&fmcount, &fmbegin, &fmend, NULL);
	printf("EOD |\n\n");
}


void print_raw_header(void)
{
	/* char *f = NULL; */
	/* char *enc = NULL; */

	/* enc = malloc(256); */
	/* if (!enc) { */
	/* 	printf("Unable to malloc 256 bytes of memory to produce dump_tape report"); */
	/* 	MHVTL_ERR("Unable to malloc 256 bytes of memory to produce dump_tape report"); */
	/* 	return; */
	/* } */
	/* f = malloc(256); */
	/* if (!f) { */
	/* 	printf("Unable to malloc 256 bytes of memory to produce dump_tape report"); */
	/* 	MHVTL_ERR("Unable to malloc 256 bytes of memory to produce dump_tape report"); */
	/* 	free(enc); */
	/* 	return; */
	/* } */

	/* sprintf(f, "%s", "Hdr:"); */
	int i;
	printf("Hdr:");
	
	switch (raw_pos.hdr.blk_type) {
	case B_DATA:
		/* if (raw_pos.hdr.blk_flags & BLKHDR_FLG_ENCRYPTED) { */
		/* 	strncat(f, "Encrypt/", 9); */
		/* } */
		/* if (raw_pos.hdr.blk_flags & BLKHDR_FLG_ZLIB_COMPRESSED) { */
		/* 	strncat(f, "zlibCompressed", 15); */
		/* } else if (raw_pos.hdr.blk_flags & BLKHDR_FLG_LZO_COMPRESSED) { */
		/* 	strncat(f, "lzoCompressed", 14); */
		/* } else { */
		/* 	strncat(f, "non-compressed", 15); */
		/* } */

		/* if (raw_pos.hdr.blk_flags & BLKHDR_FLG_CRC) { */
		/* 	strncat(f, " with crc", 10); */
		/* } else { */
		/* 	strncat(f, " no crc", 10); */
		/* } */
		if ((raw_pos.hdr.blk_flags &
			(BLKHDR_FLG_LZO_COMPRESSED | BLKHDR_FLG_ENCRYPTED)) ==
			(BLKHDR_FLG_LZO_COMPRESSED | BLKHDR_FLG_ENCRYPTED))
			printf("  Encrypt/Comp data");
		else if ((raw_pos.hdr.blk_flags &
			(BLKHDR_FLG_ZLIB_COMPRESSED | BLKHDR_FLG_ENCRYPTED)) ==
			(BLKHDR_FLG_ZLIB_COMPRESSED | BLKHDR_FLG_ENCRYPTED))
			printf("  Encrypt/Comp data");
		else if (raw_pos.hdr.blk_flags & BLKHDR_FLG_ENCRYPTED)
			printf("     Encrypted data");
		else if (raw_pos.hdr.blk_flags & BLKHDR_FLG_ZLIB_COMPRESSED)
			printf("zlibCompressed data");
		else if (raw_pos.hdr.blk_flags & BLKHDR_FLG_LZO_COMPRESSED)
			printf(" lzoCompressed data");
			else
		printf("              data");
		printf("%s", (raw_pos.hdr.blk_flags & BLKHDR_FLG_UNCOMPRESSED_CRC) ? " with crc " : " no crc  ");

		printf("(%02x), sz %6d/%-6d, Blk No.: %u, data %" PRId64 ", CRC: %08x\n",
			raw_pos.hdr.blk_type,
			raw_pos.hdr.disk_blk_size,
			raw_pos.hdr.blk_size,
			raw_pos.hdr.blk_number,
			raw_pos.data_offset,
			raw_pos.hdr.uncomp_crc);
		if (raw_pos.hdr.blk_flags & BLKHDR_FLG_ENCRYPTED) {
			printf("   => Encr key length %d, ukad length %d, "
				"akad length %d\n",
				raw_pos.hdr.blk_encryption_info.key_length,
				raw_pos.hdr.blk_encryption_info.ukad_length,
				raw_pos.hdr.blk_encryption_info.akad_length);
			printf("       Key  : ");
			for (i = 0; i < raw_pos.hdr.blk_encryption_info.key_length; i++)
				printf("%02x", raw_pos.hdr.blk_encryption_info.key[i]);
			if (raw_pos.hdr.blk_encryption_info.ukad_length > 0) {
				printf("\n       Ukad : ");
				for (i = 0; i < raw_pos.hdr.blk_encryption_info.ukad_length; i++)
					printf("%02x", raw_pos.hdr.blk_encryption_info.ukad[i]);
			}
			if (raw_pos.hdr.blk_encryption_info.akad_length > 0) {
				printf("\n       Akad : ");
				for (i = 0; i < raw_pos.hdr.blk_encryption_info.akad_length; i++)
					printf("%02x", raw_pos.hdr.blk_encryption_info.akad[i]);
			}
			puts("");
		}
		break;
	case B_FILEMARK:
		/* strncat(f, "Filemark", 9); */
		printf("         Filemark");
		printf("(%02x), sz %13d, Blk No.: %u, data %" PRId64 "\n",
			raw_pos.hdr.blk_type,
			raw_pos.hdr.blk_size,
			raw_pos.hdr.blk_number,
			raw_pos.data_offset);
		break;
	case B_EOD:
		/* strncat(f, "End of Data", 12); */
		printf("      End of Data");
		printf("(%02x), sz %13d, Blk No.: %u, data %" PRId64 "\n",
			raw_pos.hdr.blk_type,
			raw_pos.hdr.blk_size,
			raw_pos.hdr.blk_number,
			raw_pos.data_offset);
		break;
	case B_NOOP:
		/* strncat(f, "No Operation", 13); */
		printf("      No Operation");	  
		break;
	default:
		/* strncat(f, "Unknown type", 13); */
		printf("      Unknown type");
		printf("(%02x), %6d/%-6d, Blk No.: %u, data %" PRId64 "\n",
			raw_pos.hdr.blk_type,
			raw_pos.hdr.disk_blk_size,
			raw_pos.hdr.blk_size,
			raw_pos.hdr.blk_number,
			raw_pos.data_offset);
		break;
	}
	/* printf("%-35s (0x%02x/0x%02x), sz: %6d/%-6d, Blk No.: %7u, data: %10" PRId64 ", CRC: %08x\n", */
	/* 		f, */
	/* 		raw_pos.hdr.blk_type, */
	/* 		raw_pos.hdr.blk_flags, */
	/* 		raw_pos.hdr.disk_blk_size, */
	/* 		raw_pos.hdr.blk_size, */
	/* 		raw_pos.hdr.blk_number, */
	/* 		raw_pos.data_offset, */
	/* 		raw_pos.hdr.uncomp_crc); */
	/* if ((raw_pos.hdr.blk_type == B_DATA) && (raw_pos.hdr.blk_flags & BLKHDR_FLG_ENCRYPTED)) { */

	/* 	printf("   => Encr key length %d, ukad length %d, " */
	/* 			"akad length %d\n", */
	/* 			raw_pos.hdr.blk_encryption_info.key_length, */
	/* 			raw_pos.hdr.blk_encryption_info.ukad_length, */
	/* 			raw_pos.hdr.blk_encryption_info.akad_length); */

	/* 	if (raw_pos.hdr.blk_encryption_info.key_length > 0) { */
	/* 			enc_key_to_string(enc, raw_pos.hdr.blk_encryption_info.key, raw_pos.hdr.blk_encryption_info.key_length); */
	/* 			printf("%12s : %32s\n", "Key", enc); */
	/* 	} */
	/* 	if (raw_pos.hdr.blk_encryption_info.ukad_length > 0) { */
	/* 			enc_key_to_string(enc, raw_pos.hdr.blk_encryption_info.ukad, raw_pos.hdr.blk_encryption_info.ukad_length); */
	/* 			printf("%12s : %32s\n", "Ukad", enc); */
	/* 	} */
	/* 	if (raw_pos.hdr.blk_encryption_info.akad_length > 0) { */
	/* 			enc_key_to_string(enc, raw_pos.hdr.blk_encryption_info.akad, raw_pos.hdr.blk_encryption_info.akad_length); */
	/* 			printf("%12s : %32s\n", "Akad", enc); */
	/* 	} */
	/* } */

	/* free(enc); */
	/* free(f); */
}

void print_filemark_count(void)
{
	printf("Total num of filemarks: %d\n", meta.filemark_count);
}

void print_metadata(void)
{
	unsigned int a;

	for (a = 0; a < meta.filemark_count; a++)
		printf("Filemark: %d\n", filemarks[a]);
}

/*
 * Cleanup entry point
 */
void cart_deinit(void)
{
	uint32_t blk_number;

	blk_number = raw_pos.hdr.blk_number;
	MHVTL_DBG(3, "cart_deinit: flush: ibuf.idx %d, ibuf.data_written %ld, "
		  "blk_number %d", ibuf.idx, ibuf.data_written, blk_number);
	flush_index(NULL);
	cartclean();
	if (filemarks)
	  free(filemarks);
}
